package com.hncy.auth.inject.spring.boot.starter.common.extractor;

import com.hncy.auth.inject.spring.boot.starter.util.RequestUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;
import org.springframework.util.StringUtils;

import javax.servlet.http.HttpServletRequest;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.nio.charset.StandardCharsets;
import java.util.Optional;

/**
 * @Author: ying.dong
 * @Date: 2021/7/16 14:51
 */
public abstract class BaseAuthInfoExtractor<T> implements AuthInfoExtractor<T> {

    protected final Logger logger = LoggerFactory.getLogger(getClass());

    protected final String authHeader;

    public BaseAuthInfoExtractor(@NonNull String authHeader) {
        this.authHeader = authHeader;
    }

    @Override
    public Optional<T> extract() {
        HttpServletRequest httpServletRequest = RequestUtil.getRequest().orElse(null);
        if (httpServletRequest == null) {
            return Optional.empty();
        }
        return extract(httpServletRequest);
    }

    @Override
    public Optional<T> extract(HttpServletRequest request) {
        String header = request.getHeader(authHeader);
        if (StringUtils.isEmpty(header)) {
            logger.debug("cannot find any authentication info in header. {}", request.getRequestURI());
            //give a chance to find from request attribute
            header = Optional.ofNullable(request.getAttribute(authHeader))
                    .map(Object::toString)
                    .orElse(null);
        }

        if (StringUtils.isEmpty(header)) {
            logger.warn("cannot find any authentication info in request attribute. {}", request.getRequestURI());
            return Optional.empty();
        }
        String decodedAuthInfo = decodeAuthInfo(header);
        if (StringUtils.isEmpty(decodedAuthInfo)) {
            return Optional.empty();
        }
        T authInfo = doExtract(decodedAuthInfo).orElse(null);
        if (authInfo == null) {
            logger.warn("cannot extract auth info with decodedAuthInfo: {}", decodedAuthInfo);
        }
        return enhanceAuthInfo(authInfo, decodedAuthInfo);
    }

    /**
     * 解码处理
     * @param originalHeaderValue
     * @return
     */
    protected String decodeAuthInfo(String originalHeaderValue) {
        String decodedAuthInfo;
        try {
            decodedAuthInfo = URLDecoder.decode(originalHeaderValue, StandardCharsets.UTF_8.name());
        } catch (UnsupportedEncodingException e) {
            logger.warn("invalid auth info. [{}]", originalHeaderValue, e);
            return null;
        }
        return decodedAuthInfo;
    }

    /**
     * 转换
     * @param decodedAuthInfo
     * @return
     */
    protected abstract Optional<T> doExtract(String decodedAuthInfo);

    /**
     * 增强
     * @param authUser
     * @param decodedAuthInfo
     * @return
     */
    protected abstract Optional<T> enhanceAuthInfo(@Nullable T authUser, String decodedAuthInfo);
}
