package com.hncy.feign.spring.boot.starter.log;

import feign.Request;
import feign.Response;
import feign.Util;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;

import static feign.Util.*;

/**
 * @Author: ying.dong
 * @Date: 2021/4/15 09:28
 */
public class FeignSlf4jLogger extends feign.Logger {

    private final Logger logger;

    public FeignSlf4jLogger(Class<?> clazz) {
        this.logger = LoggerFactory.getLogger(clazz);
    }

    public FeignSlf4jLogger(String name) {
        this.logger = LoggerFactory.getLogger(name);
    }

    @Override
    protected void logRequest(String configKey, Level logLevel, Request request) {
        log(org.slf4j.event.Level.INFO, configKey, "---> %s %s HTTP/1.1", request.httpMethod().name(), request.url());
        if (logLevel.ordinal() >= Level.HEADERS.ordinal()) {

            for (String field : request.headers().keySet()) {
                for (String value : valuesOrEmpty(request.headers(), field)) {
                    log(org.slf4j.event.Level.DEBUG, configKey, "%s: %s", field, value);
                }
            }

            int bodyLength = 0;
            if (request.body() != null) {
                bodyLength = request.body().length;
                if (logLevel.ordinal() >= Level.FULL.ordinal()) {
                    String
                            bodyText =
                            request.charset() != null ? new String(request.body(), request.charset()) : null;
                    log(org.slf4j.event.Level.DEBUG, configKey, ""); // CRLF
                    log(org.slf4j.event.Level.DEBUG, configKey, "%s", bodyText != null ? bodyText : "Binary data");
                }
            }
            log(org.slf4j.event.Level.DEBUG, configKey, "---> END HTTP (%s-byte body)", bodyLength);
        }
    }

    @Override
    protected void logRetry(String configKey, Level logLevel) {
        log(org.slf4j.event.Level.WARN, configKey, "---> RETRYING");
    }

    @Override
    protected Response logAndRebufferResponse(String configKey, Level logLevel, Response response, long elapsedTime) throws IOException {
        String reason = response.reason() != null && logLevel.compareTo(Level.NONE) > 0 ?
                " " + response.reason() : "";
        int status = response.status();
        log(org.slf4j.event.Level.INFO, configKey, "<--- HTTP/1.1 %s%s (%sms)", status, reason, elapsedTime);
        if (logLevel.ordinal() >= Level.HEADERS.ordinal()) {

            for (String field : response.headers().keySet()) {
                for (String value : valuesOrEmpty(response.headers(), field)) {
                    log(org.slf4j.event.Level.DEBUG, configKey, "%s: %s", field, value);
                }
            }

            int bodyLength = 0;
            if (response.body() != null && !(status == 204 || status == 205)) {
                // HTTP 204 No Content "...response MUST NOT include a message-body"
                // HTTP 205 Reset Content "...response MUST NOT include an entity"
                if (logLevel.ordinal() >= Level.FULL.ordinal()) {
                    log(org.slf4j.event.Level.DEBUG, configKey, ""); // CRLF
                }
                byte[] bodyData = Util.toByteArray(response.body().asInputStream());
                bodyLength = bodyData.length;
                if (logLevel.ordinal() >= Level.FULL.ordinal() && bodyLength > 0) {
                    log(org.slf4j.event.Level.INFO, configKey, "%s", decodeOrDefault(bodyData, UTF_8, "Binary data"));
                }
                log(org.slf4j.event.Level.DEBUG, configKey, "<--- END HTTP (%s-byte body)", bodyLength);
                return response.toBuilder().body(bodyData).build();
            } else {
                log(org.slf4j.event.Level.DEBUG, configKey, "<--- END HTTP (%s-byte body)", bodyLength);
            }
        }
        return response;
    }

    @Override
    protected IOException logIOException(String configKey, Level logLevel, IOException ioe, long elapsedTime) {
        log(org.slf4j.event.Level.ERROR, configKey, "<--- ERROR %s: %s (%sms)", ioe.getClass().getSimpleName(), ioe.getMessage(),
                elapsedTime);
        if (logLevel.ordinal() >= Level.FULL.ordinal()) {
            StringWriter sw = new StringWriter();
            ioe.printStackTrace(new PrintWriter(sw));
            log(org.slf4j.event.Level.ERROR, configKey, "%s", sw.toString());
            log(org.slf4j.event.Level.ERROR, configKey, "<--- END ERROR");
        }
        return ioe;
    }

    @Override
    protected void log(String configKey, String format, Object... args) {
        //default info
        log(org.slf4j.event.Level.INFO, configKey, format, args);
    }

    protected void log(org.slf4j.event.Level frameworkLevel, String configKey, String format, Object... args) {
        if (frameworkLevel == org.slf4j.event.Level.DEBUG && logger.isDebugEnabled()) {
            logger.debug(String.format(methodTag(configKey) + format, args));
        } else if (frameworkLevel == org.slf4j.event.Level.INFO && logger.isInfoEnabled()) {
            logger.info(String.format(methodTag(configKey) + format, args));
        } else if (frameworkLevel == org.slf4j.event.Level.WARN && logger.isWarnEnabled()) {
            logger.warn(String.format(methodTag(configKey) + format, args));
        } else if (frameworkLevel == org.slf4j.event.Level.ERROR && logger.isErrorEnabled()) {
            logger.error(String.format(methodTag(configKey) + format, args));
        }
    }
}
