package com.hncy.file.manage.spring.boot.starter.oss.manager;

import cn.hutool.core.date.DateTime;
import cn.hutool.core.date.DateUtil;
import com.aliyun.oss.HttpMethod;
import com.aliyun.oss.model.GeneratePresignedUrlRequest;
import com.aliyun.oss.model.VoidResult;
import com.hncy.file.manage.spring.boot.starter.core.exception.AccessRoleNotConfiguredException;
import com.hncy.file.manage.spring.boot.starter.core.exception.ResourceException;
import com.hncy.file.manage.spring.boot.starter.core.exception.SignedException;
import com.hncy.file.manage.spring.boot.starter.oss.common.OSSClient;
import com.hncy.file.manage.spring.boot.starter.oss.common.OSSClientRegistry;
import com.hncy.file.manage.spring.boot.starter.oss.constant.OSSConstant;
import com.hncy.file.manage.spring.boot.starter.oss.properties.OSSRoleProperties;
import com.hncy.file.manage.spring.boot.starter.oss.resource.OSSResource;
import lombok.extern.slf4j.Slf4j;
import org.springframework.util.CollectionUtils;
import org.springframework.util.StringUtils;

import java.net.URL;
import java.util.*;

/**
 * 签名模式
 * @Author: ying.dong
 * @Date: 2021/5/24 19:38
 */
@Slf4j
public class OSSSignedManager extends OSSManager{

    public OSSSignedManager(OSSClientRegistry ossClientRegistry) {
        super(ossClientRegistry);
    }

    @Override
    public OSSResource get(String role, OSSResource resource) throws ResourceException {
        return doSign(role, resource, HttpMethod.GET, null);
    }

    @Override
    public OSSResource save(String role, OSSResource resource, boolean allowOverride) throws ResourceException {
        if (StringUtils.isEmpty(resource.getMD5())) {
            throw new SignedException("MD5 must not be null");
        }
        Map<String, String> headers = new HashMap<>();
        if (!allowOverride) {
            headers.put(OSSConstant.Headers.FORBID_OVERWRITE, "true");
        }
        headers.put(OSSConstant.Headers.CONTENT_MD5, resource.getMD5());
        headers.put(OSSConstant.Headers.CONTENT_TYPE, OSSConstant.Headers.CONTENT_TYPE_STREAM);
        return doSign(role, resource, HttpMethod.PUT, headers);
    }

    @Override
    public boolean del(String role, OSSResource resource) throws ResourceException {
        OSSClient ossClient = getOssClientRegistry().get(role)
                .orElseThrow(() -> new AccessRoleNotConfiguredException(role));
        OSSRoleProperties roleProperties = ossClient.getConfiguration();

        permissionValidated(role, roleProperties, HttpMethod.DELETE);

        String bucket = roleProperties.getBucket();
        VoidResult result = ossClient.getSDKClient().deleteObject(bucket, getFileName(roleProperties, resource));
        if (!result.getResponse().isSuccessful()) {
            log.error("delete resource error, resource[{}], result: {}", resource, result);
            return false;
        }
        return true;
    }

    /**
     * 签名流程
     * @param role
     * @param resource
     * @param httpMethod
     * @param additionalHeaders
     * @return
     */
    private OSSResource doSign(String role, OSSResource resource, HttpMethod httpMethod, Map<String, String> additionalHeaders) {
        OSSClient ossClient = getOssClientRegistry().get(role)
                .orElseThrow(() -> new AccessRoleNotConfiguredException(role));
        OSSRoleProperties roleProperties = ossClient.getConfiguration();

        permissionValidated(role, roleProperties, httpMethod);

        String bucket = roleProperties.getBucket();
        Map<String, String> headers = resource.getHeaders();

        if (!CollectionUtils.isEmpty(additionalHeaders)) {
            headers = CollectionUtils.isEmpty(headers) ? new HashMap<>() : headers;
            headers.putAll(additionalHeaders);
        }

        URL url = sign(roleProperties, resource, httpMethod, headers, ossClient);

        resource.setBucket(bucket);
        resource.setUrl(url);
        resource.setDirectory(getPath(roleProperties.getDirectory(), resource.getDirectory(), null));
        if (!CollectionUtils.isEmpty(headers)) {
            resource.setHeaders(headers);
        }
        return resource;
    }


    /**
     * 签名
     * @param roleProperties
     * @param resource
     * @param httpMethod
     * @param headers
     * @param ossClient
     * @return
     */
    private URL sign(OSSRoleProperties roleProperties, OSSResource resource, HttpMethod httpMethod,
                     Map<String, String> headers, OSSClient ossClient) {
        //oss directory + custom directory + name
        String fileName = getFileName(roleProperties, resource);

        GeneratePresignedUrlRequest request = new GeneratePresignedUrlRequest(roleProperties.getBucket(), fileName, httpMethod);
        DateTime dateTime = DateUtil.offsetSecond(new Date(), roleProperties.getExpiredInTimeSeconds());
        request.setExpiration(dateTime);
        request.setProcess(resource.getStyle());

        if (!CollectionUtils.isEmpty(headers)) {
            headers.forEach(request::addHeader);
        }

        try {
            return ossClient.getSDKClient().generatePresignedUrl(request);
        } catch (Exception e) {
            throw new SignedException(e);
        }
    }
}
