package com.hncy.file.manage.spring.boot.starter.oss.sdk;

import cn.hutool.core.util.ReflectUtil;
import com.aliyun.oss.ClientConfiguration;
import com.aliyun.oss.ClientException;
import com.aliyun.oss.HttpMethod;
import com.aliyun.oss.OSSClient;
import com.aliyun.oss.common.auth.Credentials;
import com.aliyun.oss.common.auth.CredentialsProvider;
import com.aliyun.oss.common.auth.ServiceSignature;
import com.aliyun.oss.common.comm.*;
import com.aliyun.oss.common.utils.HttpHeaders;
import com.aliyun.oss.common.utils.HttpUtil;
import com.aliyun.oss.internal.*;
import com.aliyun.oss.model.GeneratePresignedUrlRequest;
import org.springframework.lang.NonNull;

import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;

import static com.aliyun.oss.common.utils.CodingUtils.assertParameterNotNull;
import static com.aliyun.oss.internal.OSSConstants.DEFAULT_CHARSET_NAME;
import static com.aliyun.oss.internal.OSSUtils.*;
import static com.aliyun.oss.internal.RequestParameters.*;

/**
 * fix bug
 * (启用自定义域名时： sldEnabled == true, 生成的签名URL多包含了bucket路径，导致签名不正确 （endpoint/bucket/xxx）)
 * @Author: ying.dong
 * @Date: 2021/7/13 09:37
 */
public class BugFixedOSSClient extends OSSClient {

    public BugFixedOSSClient(@NonNull String endpoint, @NonNull CredentialsProvider credsProvider,
                             @NonNull ClientConfiguration config, @NonNull ServiceClient serviceClient) {
        super(endpoint, credsProvider, config);

        reinit(serviceClient, endpoint);
    }

    /**
     * 反射重新初始化 (fix非签名模式bug)
     * @param serviceClient
     */
    private void reinit(ServiceClient serviceClient, String endpoint) {
        ReflectUtil.setFieldValue(this, "serviceClient", serviceClient);
        ReflectUtil.invoke(this, "initOperations");
        setEndpoint(endpoint);
    }

    /**
     * fix 签名模式bug
     * @param request
     * @return
     * @throws ClientException
     */
    @Override
    public URL generatePresignedUrl(GeneratePresignedUrlRequest request) throws ClientException {
        assertParameterNotNull(request, "request");

        if (request.getBucketName() == null) {
            throw new IllegalArgumentException(OSS_RESOURCE_MANAGER.getString("MustSetBucketName"));
        }
        ensureBucketNameValid(request.getBucketName());

        if (request.getExpiration() == null) {
            throw new IllegalArgumentException(OSS_RESOURCE_MANAGER.getString("MustSetExpiration"));
        }
        String url;

        if (getClientConfiguration().getSignatureVersion() != null && getClientConfiguration().getSignatureVersion() == SignVersion.V2) {
            url = SignV2Utils.buildSignedURL(request, getCredentialsProvider().getCredentials(), getClientConfiguration(), getEndpoint());
        } else {
            url = buildSignedURL(request, getCredentialsProvider().getCredentials(), getClientConfiguration(), getEndpoint());
        }

        try {
            return new URL(url);
        } catch (MalformedURLException e) {
            throw new ClientException(e);
        }
    }


    /**
     * 签名，修复bug
     * @param request
     * @param currentCreds
     * @param config
     * @param endpoint
     * @return
     */
    private String buildSignedURL(GeneratePresignedUrlRequest request, Credentials currentCreds, ClientConfiguration config, URI endpoint) {
        String bucketName = request.getBucketName();
        String accessId = currentCreds.getAccessKeyId();
        String accessKey = currentCreds.getSecretAccessKey();
        boolean useSecurityToken = currentCreds.useSecurityToken();
        HttpMethod method = request.getMethod() != null ? request.getMethod() : HttpMethod.GET;

        String expires = String.valueOf(request.getExpiration().getTime() / 1000L);
        String key = request.getKey();
        String resourcePath = OSSUtils.determineResourcePath(bucketName, key, config.isSLDEnabled());

        RequestMessage requestMessage = new RequestMessage(bucketName, key);
        requestMessage.setEndpoint(OSSUtils.determineFinalEndpoint(endpoint, bucketName, config));
        requestMessage.setMethod(method);
        requestMessage.setResourcePath(resourcePath);
        requestMessage.setHeaders(request.getHeaders());

        requestMessage.addHeader(HttpHeaders.DATE, expires);
        if (request.getContentType() != null && !request.getContentType().trim().equals("")) {
            requestMessage.addHeader(HttpHeaders.CONTENT_TYPE, request.getContentType());
        }
        if (request.getContentMD5() != null && !request.getContentMD5().trim().equals("")) {
            requestMessage.addHeader(HttpHeaders.CONTENT_MD5, request.getContentMD5());
        }
        for (Map.Entry<String, String> h : request.getUserMetadata().entrySet()) {
            requestMessage.addHeader(OSSHeaders.OSS_USER_METADATA_PREFIX + h.getKey(), h.getValue());
        }

        Map<String, String> responseHeaderParams = new HashMap<String, String>();
        populateResponseHeaderParameters(responseHeaderParams, request.getResponseHeaders());
        if (request.getTrafficLimit() > 0) {
            responseHeaderParams.put(OSS_TRAFFIC_LIMIT, String.valueOf(request.getTrafficLimit()));
        }
        if (responseHeaderParams.size() > 0) {
            requestMessage.setParameters(responseHeaderParams);
        }

        if (request.getQueryParameter() != null && request.getQueryParameter().size() > 0) {
            for (Map.Entry<String, String> entry : request.getQueryParameter().entrySet()) {
                requestMessage.addParameter(entry.getKey(), entry.getValue());
            }
        }

        if (request.getProcess() != null && !request.getProcess().trim().equals("")) {
            requestMessage.addParameter(RequestParameters.SUBRESOURCE_PROCESS, request.getProcess());
        }

        if (useSecurityToken) {
            requestMessage.addParameter(SECURITY_TOKEN, currentCreds.getSecurityToken());
        }

        String canonicalResource = "/" + ((bucketName != null) ? bucketName : "") + ((key != null ? "/" + key : ""));
        String canonicalString = SignUtils.buildCanonicalString(method.toString(), canonicalResource, requestMessage,
                expires);
        String signature = ServiceSignature.create().computeSignature(accessKey, canonicalString);

        Map<String, String> params = new LinkedHashMap<String, String>();
        params.put(HttpHeaders.EXPIRES, expires);
        params.put(OSS_ACCESS_KEY_ID, accessId);
        params.put(SIGNATURE, signature);
        params.putAll(requestMessage.getParameters());

        String queryString = HttpUtil.paramToQueryString(params, DEFAULT_CHARSET_NAME);

        /* Compse HTTP request uri. */
        String url = requestMessage.getEndpoint().toString();
        if (!url.endsWith("/")) {
            url += "/";
        }

        //fix bug
        String finalResourcePath = resourcePath;
        if (config.isSLDEnabled()) {
            finalResourcePath = OSSUtils.makeResourcePath(key);
        }
        url += finalResourcePath + "?" + queryString;
        return url;
    }
}
