package com.hncy.message.spring.boot.starter.consumer;

import com.hncy.message.spring.boot.starter.metadata.MessageMetadata;
import com.hncy.message.spring.boot.starter.metadata.resolver.MessageMetadataResolver;
import com.rabbitmq.client.Channel;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.amqp.core.Message;
import org.springframework.amqp.rabbit.listener.adapter.AbstractAdaptableMessageListener;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.lang.NonNull;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Consumer;

/**
 * @Author: ying.dong
 * @Date: 2021/6/3 16:32
 */
public abstract class DynamicMessageListener<T> extends AbstractAdaptableMessageListener implements InitializingBean {

    protected final MessageMetadata messageMetadata;

    protected final MessageMetadataResolver messageMetadataResolver;

    /**
     * 最大重试3次
     */
    private final int DEFAULT_MAX_RETRIES = 3;

    private int maxRetries = DEFAULT_MAX_RETRIES;

    private Map<String, AtomicInteger> retryCache = new ConcurrentHashMap<>();

    protected final Logger logger = LoggerFactory.getLogger(getClass());

    /**
     * 是否自动创建MQ资源
     */
    protected final boolean autoCreateMQ;

    public DynamicMessageListener(@NonNull MessageMetadata messageMetadata, @NonNull MessageMetadataResolver messageMetadataResolver) {
        this(messageMetadata, messageMetadataResolver, true);
    }

    public DynamicMessageListener(@NonNull MessageMetadata messageMetadata, @NonNull MessageMetadataResolver messageMetadataResolver, boolean autoCreateMQ) {
        this.messageMetadata = messageMetadata;
        this.messageMetadataResolver = messageMetadataResolver;
        this.autoCreateMQ = autoCreateMQ;
    }

    @Override
    public void afterPropertiesSet() throws Exception {
        logger.debug("autoCreateMQ configured with [{}]", autoCreateMQ);
        if (autoCreateMQ) {
            messageMetadataResolver.resolve(messageMetadata);
            logger.info("MQ auto declared with metadata [{}]", messageMetadata);
        }
    }


    /**
     * 处理完毕，正常回应，此消息将被删除
     * @param message
     * @param channel
     * @throws IOException
     */
    protected void ackOK(Message message, Channel channel) throws IOException {
        channel.basicAck(message.getMessageProperties().getDeliveryTag(), false);
        logger.info("message [{}] ack: ok.", message);
    }

    /**
     * 消息将再次入队
     * @param message
     * @param channel
     */
    protected void ackEnqueue(Message message, Channel channel) throws IOException {
        channel.basicReject(message.getMessageProperties().getDeliveryTag(), true);
        logger.info("message [{}] ack: enqueue.", message);
    }

    /**
     * 丢弃该条消息
     * @param message
     * @param channel
     * @throws IOException
     */
    protected void ackDiscard(Message message, Channel channel) throws IOException {
        channel.basicReject(message.getMessageProperties().getDeliveryTag(), false);
        logger.warn("message [{}] ack: discard.", message);
    }

    /**
     * 失败时处理
     * @param message
     * @param channel
     * @return
     */
    public boolean onError(Message message, Channel channel) throws IOException {
        ackEnqueue(message, channel);
        return true;
    }

    /**
     * 达到最大重试时处理
     * @param message
     * @param channel
     * @return
     * @throws IOException
     */
    public boolean onMaxRetries(Message message, Channel channel) throws IOException {
        ackDiscard(message, channel);
        return true;
    }

    /**
     * 执行模板
     * @param message
     * @param channel
     * @param consumer
     * @throws IOException
     */
    protected final void doExecute(Message message, Channel channel, Consumer<Object> consumer) throws IOException {
        boolean alreadyAck = false;
        String messageId = message.getMessageProperties().getMessageId();
        try {
            if (!retryCache.containsKey(messageId)) {
                retryCache.put(messageId, new AtomicInteger(0));
            } else {
                retryCache.get(messageId).addAndGet(1);
            }
            Object object = getMessageConverter().fromMessage(message);
            consumer.accept(object);
        } catch (Exception e) {
            logger.error("consume message error.", e);
            if (retryCache.get(messageId).get() > maxRetries) {
                logger.warn("consume message retry limited. [{}]", message);
                alreadyAck = onMaxRetries(message, channel);
                return;
            }
            alreadyAck = onError(message, channel);
        } finally {
            if (!alreadyAck) {
                ackOK(message, channel);
                retryCache.remove(messageId);
            }
        }
    }

    public MessageMetadata getMessageMetadata() {
        return messageMetadata;
    }

    public void setMaxRetries(int maxRetries) {
        this.maxRetries = maxRetries;
    }
}
