package com.hncy.sms.spring.boot.starter.aliyun.request;

import org.springframework.util.StringUtils;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

/**
 * 阿里云短信参数构造器
 * @Author: ying.dong
 * @Date: 2021/5/12 17:28
 */
public class AliyunSmsSendRequestBuilder {

    /**
     * 接收短信的手机号码。
     *
     * 格式：
     *
     * 国内短信：+/+86/0086/86或无任何前缀的11位手机号码，例如1381111****。
     * 国际/港澳台消息：国际区号+号码，例如852000012****。
     * 上限为1000个手机号码。批量调用相对于单条调用及时性稍有延迟。
     * 必填
     */
    private Collection<String> phoneNumbers = new HashSet<>();

    /**
     * 短信签名名称。请在控制台国内消息或国际/港澳台消息页面中的签名管理页签下签名名称一列查看。
     * 必填
     */
    private String signName;

    /**
     * 短信模板ID。请在控制台国内消息或国际/港澳台消息页面中的模板管理页签下模板CODE一列查看。
     * 必填
     */
    private String templateCode;

    /**
     * 短信模板变量对应的实际值
     * 非必填
     */
    private Map<String, Object> templateParam = new HashMap<>();

    /**
     * 上行短信扩展码，上行短信，指发送给通信服务提供商的短信，用于定制某种服务、完成查询，或是办理某种业务等，需要收费的，按运营商普通短信资费进行扣费。
     * 非必填
     */
    private String smsUpExtendCode;

    /**
     * 外部流水扩展字段
     * 非必填
     */
    private String outId;

    public static AliyunSmsSendRequestBuilder getInstance() {
        return new AliyunSmsSendRequestBuilder();
    }

    public AliyunSmsSendRequestBuilder phones(String phone) {
        if (StringUtils.isEmpty(phone)) {
            throw new IllegalArgumentException("phone must not be null");
        }
        this.phoneNumbers.add(phone);
        return this;
    }

    public AliyunSmsSendRequestBuilder signName(String signName) {
        this.signName = signName;
        return this;
    }

    public AliyunSmsSendRequestBuilder templateCode(String templateCode) {
        if (StringUtils.isEmpty(templateCode)) {
            throw new IllegalArgumentException("sms templateCode must not be null");
        }
        this.templateCode = templateCode;
        return this;
    }

    public AliyunSmsSendRequestBuilder templateParams(String param, String value) {
        this.templateParam.put(param, value);
        return this;
    }

    public AliyunSmsSendRequestBuilder smsUpExtendCode(String smsUpExtendCode) {
        this.smsUpExtendCode = smsUpExtendCode;
        return this;
    }

    public AliyunSmsSendRequestBuilder outId(String outId) {
        this.outId = outId;
        return this;
    }

    public AliyunSmsSendRequest build() {
        AliyunSmsSendRequest request = new AliyunSmsSendRequest();
        request.setPhoneNumbers(phoneNumbers);
        request.setSignName(signName);
        request.setTemplateCode(templateCode);
        if (!templateParam.isEmpty()) {
            request.setTemplateParam(templateParam);
        }
        request.setSmsUpExtendCode(smsUpExtendCode);
        request.setOutId(outId);
        return request;
    }
}
