package com.hncy.deploy.service;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.date.LocalDateTimeUtil;
import cn.hutool.core.io.FileUtil;
import cn.hutool.http.HttpUtil;
import com.hncy.deploy.dto.ApplicationInfo;
import com.hncy.deploy.dto.ApplicationProcess;
import com.hncy.deploy.exception.HealthCheckException;
import com.hncy.deploy.exception.JarFileNotExists;
import com.hncy.deploy.exception.ProcessNotExists;
import com.hncy.deploy.exception.ServiceNotConfiguredException;
import com.hncy.deploy.properties.DeployProperties;
import com.hncy.deploy.properties.FlatDeployProperties;
import com.hncy.deploy.properties.JvmProperties;
import com.hncy.deploy.properties.ServiceProperties;
import com.hncy.deploy.util.CommandUtil;
import com.hncy.jackson.spring.boot.starter.core.JSONObject;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.BeanUtils;

import javax.annotation.PostConstruct;
import java.io.File;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.stream.Collectors;

/**
 * @Author: ying.dong
 * @Date: 2021/5/18 10:26
 */
@Slf4j
public class DefaultDeployService implements DeployService {

    private static final String DEFAULT_LAUNCHER = "service-";

    /**
     * 默认应用启动时间
     */
    private static final int DEFAULT_APPLICATION_LAUNCH_TIME_IN_SECONDS = 20;

    /**
     * 默认应用配置文件目录
     */
    private static final String DEFAULT_CONFIG_FILE_FOLDER = "config";

    /**
     * 默认网络超时时间（秒）
     */
    private static final int DEFAULT_SOCKET_TIMEOUT_IN_SECONDS = 3;

    /**
     * 默认网络超时时间（毫秒）
     */
    private static final int DEFAULT_SOCKET_TIMEOUT_IN_MILLIS = DEFAULT_SOCKET_TIMEOUT_IN_SECONDS * 1000;


    /**
     * 健康检查成功状态
     */
    private static final String HEALTH = "UP";

    private final DeployProperties deployProperties;

    /**
     * 名称: FlatDeployProperties
     */
    private Map<String, FlatDeployProperties> nameMap;

    /**
     * jar : FlatDeployProperties
     */
    private Map<String, FlatDeployProperties> jarMap;

    /**
     * 默认执行的jar包名称
     */
    private static final String BASE_JAR_FILE_NAME = "exec.jar";

    /**
     * jar备份名称
     */
    private static final String BASE_JAR_FILE_BAK_NAME = "exec.jar.bak";

    private static final DateTimeFormatter FORMATTER = DateTimeFormatter.ofPattern("yyyyMMddHHmmss");

    public DefaultDeployService(DeployProperties deployProperties) {
        this.deployProperties = deployProperties;
    }

    @PostConstruct
    private void flatProperties() {
        String workspace = deployProperties.getWorkspace();
        String profile = deployProperties.getProfile();
        nameMap = new HashMap<>();
        jarMap = new HashMap<>();
        for (ServiceProperties it: deployProperties.getService()) {
            FlatDeployProperties flatDeployProperties = new FlatDeployProperties();
            BeanUtils.copyProperties(it, flatDeployProperties);
            flatDeployProperties.setProfile(profile);
            flatDeployProperties.setWorkspace(workspace);
            flatDeployProperties.setShutdownTimeoutInSeconds(deployProperties.getShutdownTimeoutInSeconds());
            flatDeployProperties.setStartTimeoutInSeconds(deployProperties.getStartTimeoutInSeconds());
            nameMap.put(it.getName(), flatDeployProperties);
            String jar = deployProperties.getWorkspace() + "/" + it.getDirectory();
            jarMap.put(jar, flatDeployProperties);
        }
    }

    @Override
    public List<ApplicationInfo> list() {
        return nameMap.values().stream()
                .map(it -> {
                    ApplicationInfo applicationInfo = new ApplicationInfo();
                    BeanUtils.copyProperties(it, applicationInfo);
                    return applicationInfo;
                }).collect(Collectors.toList());
    }

    @Override
    public List<ApplicationProcess> ps(String name) {
        String serviceName = Optional.ofNullable(nameMap.get(name))
                .map(FlatDeployProperties::getName)
                .orElseThrow(() -> new ServiceNotConfiguredException(name));
        return psForName(serviceName);
    }

    @Override
    public ApplicationProcess ps(Integer id) {
        return ps().stream()
                .filter(it -> it.getId().equals(id))
                .findAny()
                .orElse(null);
    }

    @Override
    public List<ApplicationProcess> ps() {
        return psForName(DEFAULT_LAUNCHER);
    }

    @Override
    public boolean isProcessExists(String name) {
        return CollectionUtil.isNotEmpty(ps(name));
    }

    @Override
    public boolean shutdownGraceful(String name) {
        List<ApplicationProcess> processes = ps(name);
        for (ApplicationProcess process: processes) {
            shutdownGraceful(process.getId());
        }
        return true;
    }

    @Override
    public boolean shutdownGraceful(Integer id) {
        if (id == null) {
            throw new ProcessNotExists(id);
        }
        int timeoutInSeconds = deployProperties.getShutdownTimeoutInSeconds();
        if (timeoutInSeconds <= 0) {
            throw new IllegalArgumentException("timeoutInSeconds must greater than zero!");
        }
        String command = "kill -15 " + id;
        String result = CommandUtil.execute(command);
        log.info("shutdown process with id [{}], result: {}{}", id, result, true);
        String prompt = ".";
        for (int i = 0; i <= timeoutInSeconds; i = i+3) {
            if (ps(id) == null) {
                return true;
            }
            log.info("waiting for shutdown application with id [{}]{}", id, prompt);
            try {
                Thread.sleep(3000);
            } catch (InterruptedException e) {
                log.error("failed to waiting for shutdown application with id [{}]", id);
            }
            prompt += ".";
        }

        log.warn("waiting for shutdown application with id [{}] timeout. application will be shutdown forced.", id);
        //force shutdown
        shutdownForce(id);
        return true;
    }

    @Override
    public boolean shutdownForce(Integer id) {
        String command = "kill -9 " + id;
        String result = CommandUtil.execute(command);
        try {
            Thread.sleep(2000);
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        log.info("force shutdown for application with id [{}], result: {}", id, result);
        return true;
    }

    @Override
    public void start(String name) {
        long begin = System.currentTimeMillis();

        FlatDeployProperties deployProperties = Optional.ofNullable(nameMap.get(name))
                .orElseThrow(() -> new ServiceNotConfiguredException(name));
        File jarFile = getJarFile(name, BASE_JAR_FILE_NAME, true);

        boolean useExternalConfigFile = true;
        File configFile = FileUtil.file(deployProperties.getWorkspace(), deployProperties.getDirectory(),
                DEFAULT_CONFIG_FILE_FOLDER);
        if (!configFile.exists() || FileUtil.isDirEmpty(configFile)) {
            log.info("config file [{}] not exists. default configuration will be used.", configFile);
            useExternalConfigFile = false;
        }
        JvmProperties jvmProperties = deployProperties.getJvm();
        StringBuilder commandBuilder = new StringBuilder("java")
                .append(" ").append(jvmProperties.getMemOpt())
                .append(" ").append(jvmProperties.getGcOpt())
                .append(" ").append(jvmProperties.getGcLogOpt())
                .append(" ").append("-jar")
                .append(" ").append(jarFile)
                .append(" ").append("--spring.profiles.active=").append(deployProperties.getProfile());
        if (useExternalConfigFile) {
            String path = configFile.toURI().getPath();
            log.info("use configuration file: {}", path);
            commandBuilder.append(" ").append("--spring.config.location=").append(path);
        }
        commandBuilder.append(" ").append("1>/dev/null 2>&1 &");
        String command = commandBuilder.toString();
        String result = CommandUtil.execute(command);
        log.info("start executed. {}", result);

        String prompt = " is starting.";
        //默认等待应用启动一定时间
        for (int i = 0; i < DEFAULT_APPLICATION_LAUNCH_TIME_IN_SECONDS; i = i + DEFAULT_SOCKET_TIMEOUT_IN_SECONDS) {
            log.info("{}{}", name, prompt);
            try {
                Thread.sleep(DEFAULT_SOCKET_TIMEOUT_IN_MILLIS);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
            prompt += ".";
        }

        long end = System.currentTimeMillis();
        int timeRemain = deployProperties.getStartTimeoutInSeconds() - (int) ((end - begin) / 1000);
        int i = 0;
        while (i < timeRemain) {
            log.info("{}{}", name, prompt);
            boolean isHealth;
            try {
                isHealth = isHealthy(name);
            } catch (Exception e) {
                isHealth = false;
                log.error("health check error. retrying ....");
            }
            if (isHealth) {
                log.info("{} started success!", name);
                return;
            }
            try {
                Thread.sleep(DEFAULT_SOCKET_TIMEOUT_IN_MILLIS);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
            prompt += ".";
            i += DEFAULT_SOCKET_TIMEOUT_IN_SECONDS;
        }
        log.warn("application maybe started failed. please check health again....");
    }

    @Override
    public void restart(String name) {
        shutdownGraceful(name);
        start(name);
    }

    @Override
    public String health(String name) {
        FlatDeployProperties flatDeployProperties = getProperties(name);
        try {
            String result = HttpUtil.get(flatDeployProperties.getHealthCheck(), DEFAULT_SOCKET_TIMEOUT_IN_MILLIS);
            Map<String, Object> jsonResult = JSONObject.getInstance().parseMap(result, String.class, Object.class);
            return (String) jsonResult.get("status");
        } catch (Exception e) {
            throw new HealthCheckException(name, e);
        }
    }

    @Override
    public boolean isHealthy(String name) {
        return HEALTH.equals(health(name));
    }

    @Override
    public void upgrade(String name) {
        FlatDeployProperties flatDeployProperties = getProperties(name);
        File jarFile = getJarFile(name, flatDeployProperties.getJar(), true);
        File execJarFile = getJarFile(name, BASE_JAR_FILE_NAME, false);
        File bakFile = getJarFile(name, BASE_JAR_FILE_BAK_NAME, false);
        if (bakFile.exists()) {
            //rename exec.bak to exec.2021xxxxxx
            FileUtil.rename(bakFile, LocalDateTimeUtil.format(LocalDateTimeUtil.now(), FORMATTER), true);
        }

        shutdownGraceful(name);

        if (execJarFile.exists()) {
            //rename exec.jar to exec.jar.bak
            FileUtil.rename(execJarFile, BASE_JAR_FILE_BAK_NAME, false);
        }

        //rename xxx.jar to exec.jar
        FileUtil.rename(jarFile, BASE_JAR_FILE_NAME, false);

        start(name);
    }

    @Override
    public void rollback(String name) {
        FlatDeployProperties flatDeployProperties = getProperties(name);
        File execJarFile = getJarFile(name, BASE_JAR_FILE_NAME, true);
        File bakFile = getJarFile(name, BASE_JAR_FILE_BAK_NAME, true);

        shutdownGraceful(name);

        //rename exec.jar to xxxx.jar
        FileUtil.rename(execJarFile, flatDeployProperties.getJar(), true);
        //rename exec.jar.bak to exec.jar
        FileUtil.rename(bakFile, BASE_JAR_FILE_NAME, false);

        start(name);
    }

    /**
     * 根据启动类查进程
     * @param name
     * @return
     */
    private List<ApplicationProcess> psForName(String name) {
        String command = "ps -ef | grep java | grep " + name + " | awk '{print $2,$0}'";
        return CommandUtil.executeLines(command).stream()
                .map(it -> {
                    String[] array = it.split(" ");
                    if (array.length < 2) {
                        return null;
                    }

                    for (Map.Entry<String, FlatDeployProperties> entry: jarMap.entrySet()) {
                        if (!it.contains(entry.getKey())) {
                            continue;
                        }
                        ApplicationProcess applicationProcess = new ApplicationProcess();
                        applicationProcess.setId(Integer.valueOf(array[0]));
                        FlatDeployProperties flatDeployProperties = entry.getValue();
                        applicationProcess.setName(flatDeployProperties.getName());
                        applicationProcess.setDescription(flatDeployProperties.getDescription());
                        applicationProcess.setDirectory(flatDeployProperties.getDirectory());
                        return applicationProcess;
                    }

                    return null;
                }).filter(Objects::nonNull)
                .collect(Collectors.toList());
    }

    /**
     * 根据名称获取配置
     * @param serviceName
     * @return
     */
    private FlatDeployProperties getProperties(String serviceName) {
        return Optional.ofNullable(nameMap.get(serviceName))
                .orElseThrow(() -> new ServiceNotConfiguredException(serviceName));
    }

    /**
     * 根据名称获取Jar文件
     * @param name
     * @param baseJarName
     * @param validate
     * @return
     */
    private File getJarFile(String name, String baseJarName, boolean validate) {
        FlatDeployProperties deployProperties = getProperties(name);
        File jarFile = FileUtil.file(deployProperties.getWorkspace(), deployProperties.getDirectory(), baseJarName);
        if (validate && !jarFile.exists()) {
            throw new JarFileNotExists(jarFile.toString());
        }
        return jarFile;
    }

}
